/* Author: earthengine@gmail.com           */
/* License: GPL 2.0                        */
#include "VDICore.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DUP(x) x,x

DECLINLINE(void)
PrintUUID(PRTUUID uuid) {
	printf("%08x-%04x-%04x-%02x-%02x-%02x%02x%02x%02x%02x%02x\n",
		uuid->Gen.u32TimeLow,uuid->Gen.u16TimeMid,uuid->Gen.u16TimeHiAndVersion,
		uuid->Gen.u8ClockSeqHiAndReserved,uuid->Gen.u8ClockSeqLow,
		uuid->Gen.au8Node[0],uuid->Gen.au8Node[1],uuid->Gen.au8Node[2],
		uuid->Gen.au8Node[3],uuid->Gen.au8Node[4],uuid->Gen.au8Node[5]);
}

DECLINLINE(void)
Error(const char *str) {
	printf("%s\n", str);
	exit(1);
}

DECLINLINE(void)
usage() {
	Error("Usage: vdireader <vdi file name>");
}

DECLINLINE(void)
OutputPreHeader(PVDIPREHEADER ppreheader) {
	printf("szFileInfo : %s\n", ppreheader->szFileInfo);
	printf("u32Signature : %08x\n", ppreheader->u32Signature);
	printf("u32Version : %d.%d\n", VDI_GET_VERSION_MAJOR(ppreheader->u32Version),
		VDI_GET_VERSION_MINOR(ppreheader->u32Version));
}

DECLINLINE(const char *)
GetImageTypeStr(VDIIMAGETYPE type) {
	switch(type) {
	case VDI_IMAGE_TYPE_NORMAL:
		return "Dynamic Growing";
	case VDI_IMAGE_TYPE_FIXED:
		return "Fixed Size";
	case VDI_IMAGE_TYPE_UNDO:
		return "Dynamic Growing for Undo/Commit";
	case VDI_IMAGE_TYPE_DIFF:
		return "Dynamic Growing for Differencing";
	default:
		return "Unknown";
	}
}

DECLINLINE(void)
OutputGeometry(PVDIDISKGEOMETRY geo) {
	if(geo==NULL)
		printf("Geometry Data Not Available\n");
	else {
		printf("C : %d(0x%x) H : %d(0x%x) S : %d(0x%x) SS : %d(0x%x) Total Size in Bytes : %lld\n"
			,DUP(geo->cCylinders) ,DUP(geo->cHeads) ,DUP(geo->cSectors)
			,DUP(geo->cbSector), 1LL * geo->cCylinders * geo->cHeads * geo->cSectors * geo->cbSector);
        }
}

DECLINLINE(void)
OutputHeader(PVDIHEADER pheader) {
	printf("Size of header : %d(0x%x)\tFlags : %d(0x%x)\tImage Type: %s\n",
		DUP(getHeaderSize(pheader)),DUP(getImageFlags(pheader)),GetImageTypeStr(getImageType(pheader)));

	printf("Comments : %s\n",getImageComment(pheader));

	printf("Block Map Offset : %d(0x%x)\tData Offset : %d(0x%x)\n",
		DUP(getImageBlocksOffset(pheader)),DUP(getImageDataOffset(pheader)));
	
	printf("Sizeof disk (in bytes) : %lld(0x%llx)\tBlock Size : %d(0x%x)\n",
		DUP((long long)getImageDiskSize(pheader)),DUP(getImageBlockSize(pheader)));

	printf("Numbe of Blocks : %d(0x%x)\tAllocated : %d(0x%x)\n",
		DUP(getImageBlocks(pheader)),DUP(getImageBlocksAllocated(pheader)));

	printf("Numbe of Extra Blocks : %d(0x%x)\n",
		DUP(getImageExtraBlockSize(pheader)));


	printf("UUID of Image : ");
	PrintUUID(getImageCreationUUID(pheader));
	printf("UUID of Image's Last Modification : ");
	PrintUUID(getImageModificationUUID(pheader));
	printf("UUID of Previous Image : ");
	PrintUUID(getImageParentUUID(pheader));
	printf("UUID of Previous Image's Last Modicication : ");
	PrintUUID(getImageParentModificationUUID(pheader));

	OutputGeometry(getImageLCHSGeometry(pheader));
}

DECLINLINE(void)
ReadAndValidateBuffer(FILE *file, PVDIHEADER header, void *pbuffer, size_t size) {
	int c=fread(pbuffer, 1, size, file);
	if(c<sizeof(header->u.v0))
		Error("Invalid File Format\n");
}

DECLINLINE(PVDIIMAGEBLOCKPOINTER)
GetBlockAllocation(FILE *file, PVDIHEADER pheader) {
	int blockscount = getImageBlocks(pheader);

	PVDIIMAGEBLOCKPOINTER BlockMap = (PVDIIMAGEBLOCKPOINTER)malloc(blockscount*sizeof(VDIIMAGEBLOCKPOINTER));
	if(!BlockMap) return NULL;
	
	fseek(file,getImageBlocksOffset(pheader),SEEK_SET);

	size_t c=fread(BlockMap, sizeof(VDIIMAGEBLOCKPOINTER), blockscount, file);
	if(c<blockscount)
		Error("Invalid File Format\n");

	return BlockMap;
}

DECLINLINE(void)
WriteOut(FILE *file, PVDIHEADER pheader, PVDIIMAGEBLOCKPOINTER BlockMap, int nowrite) {
	FILE *outfile = fopen("out.img","wb");
	if(outfile==NULL) Error("Error open out.img\n");

	uint32_t blocksize = getImageBlockSize(pheader);
	void *buffer = malloc(blocksize);
	if(buffer==NULL) {
		fclose(outfile);
		Error("Memory Allocation Failue\n");
	}

	uint32_t dataoff = getImageDataOffset(pheader);
	int i;
	for( i=0; i<getImageBlocks(pheader); ++i) {
		if(IS_VDI_IMAGE_BLOCK_ALLOCATED(BlockMap[i])) {
			putchar('*');
			if(!nowrite) {
				fseek(file, dataoff + BlockMap[i] * blocksize, SEEK_SET);
				fread(buffer, 1, blocksize, file);
			}
		}
		else if(BlockMap[i]==VDI_IMAGE_BLOCK_ZERO) {
			putchar('=');
			bzero(buffer, blocksize);
		}
		else {
			putchar('-');
		}
		if(!nowrite)
			fwrite(buffer, 1, blocksize, outfile);
	}
	putchar('\n');
	
	free(buffer);
	fclose(outfile);
}

DECLINLINE(void)
ProcessFile(FILE *file,int nowrite) {
	VDIPREHEADER preheader;
	int c=fread(&preheader,1,sizeof(preheader),file);
	if(c<sizeof(preheader) || preheader.u32Signature!=VDI_IMAGE_SIGNATURE)
		Error("Invalid File Format\n");

	OutputPreHeader(&preheader);

	VDIHEADER header;
	header.uVersion = preheader.u32Version;

	ReadAndValidateBuffer(file,&header,&header.u.v1plus,sizeof(header.u.v1plus));

	OutputHeader(&header);

	PVDIIMAGEBLOCKPOINTER BlockMap = GetBlockAllocation(file, &header);

	WriteOut(file, &header, BlockMap, nowrite);

	free(BlockMap);
}

int
main(int argc, const char *argv[]) {
	if(argc<2 || argc>3) usage();
	if(argc==3 && argv[2][0]!='r') usage();
	
	int nowrite=0;
	if(argc==3) nowrite=1;

	FILE *file = fopen(argv[1],"rb");
	if(file==NULL) Error("Cannot Open File");
	
	ProcessFile(file,nowrite);

	fclose(file);

	return 0;
}
